#include "CNode.h"
#include <iostream>

using namespace std;

template <typename T> class List;
template <typename T> std::ostream & operator<<(std::ostream &, const List<T> &);
template <typename T> bool operator*(const T &, const List<T> &);

template <typename T> class List {
private:
    Node<T> * head;
    Node<T> * tail;
    int c; // ilosc elementow w liscie
public:
    List() {
	head = tail = NULL;
	c = 0;
    }
    
    int count() const { return c; }
    
    List(const List<T> &lista) {
	head = NULL;
	tail = NULL;
	c = 0;
	int cl = lista.count();
	for (int i=0; i<cl; i++)
	    (*(this))+=(*(lista[i]));
    }
    
    ~List();
    List<T> &operator+=(const T &item); // lista + element #
    List<T> &operator+=(const List &list); // dodanie list #
    List<T> operator+(const List &list); // dodanie list #
    List<T> operator+(const T &item); // dodanie elementu #
    List<T> &operator-=(const T &item); // usuniecie elementu z listy #
    List<T> operator-(const T &item); // usuniecie elementu z listy #
    List<T> &operator-=(const List &lista); // odjecie listy #
    List<T> operator-(const List &lista); // odejmowanie list #
    T* operator[](int i) const; // wyciagniecie wskaznika do elementu o danym indeksie #
    int operator()(const T &item); // wyciagniecie indeksu podanego elementu #
    bool operator*(const T &item) const; // czy element jest na liscie #
    friend bool (::operator*<>)(const T &, const List<T> &); // czy element jest na liscie #
    friend List<T> operator+<>(const T &item, const List<T> &lista); // element + lista #
    friend ostream &operator<<<>(ostream &, const List<T> &lista); // wyslanie na stdout #
    List<T> operator^(const List<T> &lista); // czesc wspolna list #
    List<T> &operator^=(const List<T> &lista); // czesc wspolna list #
    List<T> &operator=(const List<T> lista); // przypisanie listy #
    bool operator==(const List<T> &lista); // czy listy sa takie same #
    bool operator!=(const List<T> &lista); // czy listy sa rozne #
};

template <typename T> List<T> &List<T>::operator^=(const List<T> &lista) {
    (*this)=(*this)^lista;
    return (*this);
}

template <typename T> bool operator*(const T &item, const List<T> &lista) {
    return lista*item;
}

template <typename T> int List<T>::operator()(const T &item) {
    Node<T> * node = head;
    int i = 0;
    while (node) {
	if (node->data == item) return i;
	i++;
	node = node->next;
    }
    return -1;
}

template <typename T> List<T> List<T>::operator-(const List &lista) {
    List<T> nowalista;
    nowalista = *this;
    nowalista-=lista;
    return nowalista;
}

template <typename T> List<T> &List<T>::operator-=(const List &lista) {
    int cl = lista.count();
    for (int i=0; i<cl; i++) {
	if ((*this)(*(lista[i]))>=0) *this-=(*(lista[i]));
    }
    return *this;
}

template <typename T> List<T> List<T>::operator-(const T &item) {
    List<T> nowalista;
    nowalista = (*this);
    nowalista-=item;
    return nowalista;
}

template <typename T> List<T> List<T>::operator+(const T &item) {
    List<T> nowalista;
    nowalista=(*this);
    nowalista+=item;
    return nowalista;
}

template <typename T> List<T> &List<T>::operator+=(const List &lista) {
    int cl = lista.count();
    for (int i=0; i<cl; i++)
	(*(this))+=(*(lista[i]));
    return *this;
}

template <typename T> bool List<T>::operator!=(const List<T> &lista) {
    return (!((*this) == lista));
}

template <typename T> List<T> &List<T>::operator=(const List<T> lista) {
    Node<T> * node = head;
    Node<T> * temp;
    while (node != NULL) {
	temp = node;
	node = node -> next;
	delete temp;
    }
    head = NULL;
    tail = NULL;
    c = 0;
    int cl = lista.count();
    for (int i=0; i<cl; i++)
	(*(this))+=(*(lista[i]));
    return *this;
}

template <typename T> List<T> List<T>::operator^(const List<T> &lista) {
    List<T> nowalista;
    int m = (c < lista.count() ? c : lista.count());
    for (int i=0; i<m; i++) {
	if((*(*this)[i])==(*(lista[i]))) nowalista+= (*(lista[i]));
    }
    return nowalista;
}

template <typename T> bool List<T>::operator==(const List<T> &lista) {
    if (c == lista.count()) {
	for (int i=0; i<c; i++) {
	    if((*(*this)[i])!=(*(lista[i]))) return false;
	}
	return true;
    }
    return false;
}

template <typename T> ostream &operator<<(ostream &ekran, const List<T> &lista) {
    Node<T> * node = lista.head;
    while (node) {
	cout << node->data << endl;
	node = node->next;
    }
    return ekran;
}

template <typename T> List<T> operator+(const T &item, const List<T> &lista) {
    List<T> nowalista;
    nowalista = lista;
    nowalista += item;
    return nowalista;
}
    
template <typename T> List<T> &List<T>::operator-=(const T &item){ // usuniecie elementu z listy
    Node<T> * node = head;
    while (node != NULL) {
	if (node->data == item) { // szukamy elementu do usuniecia
	    if (node->prev) node->prev->next = node -> next; else head = node -> next;
	    if (node->next) node->next->prev = node->prev; else tail = node->prev;
	    delete node; // usuniecie elementu z pamieci
	    c--;
	    break;
	}
	node = node -> next;
    }
    return *this;
}
    
template <typename T> bool List<T>::operator*(const T &item) const { // operator sprawdzajacy, czy dany element jest w liscie
    Node<T> * node = head;
    while (node != NULL) {
	if (node->data == item) {
	    return true;
	}
	node = node -> next;
    }
    return false;
}
    
template <typename T> T* List<T>::operator[](int i) const { // zwraca element o danym indeksie
    if (i<c) {
        int j = 0;
        Node<T> * node = head;
        for (j=0; j<i; j++)
	    if (node) node = node -> next;
	    return &(node->data);
	}
	return NULL;
    }
    
template <typename T> List<T> List<T>::operator+(const List &list) { // laczenie list w nowa
	List<T> newl;
	for (int i=0; i<c; i++)
	    newl+=(*(*this)[i]);
	int m = list.count();
	for (int i=0; i<m; i++) {
	    newl+=(*(list[i]));
	}
	return newl;
    }
    
template <typename T> List<T>::~List() { // usuniecie elementow z pamieci
	Node<T> * node = head;
	Node<T> * temp;
	while (node != NULL) {
	    temp = node;
	    delete temp;
	    node = node -> next;
	}	
    }

template <typename T> List<T>& List<T>::operator+=(const T &item) { // dodawanie elementow do listy
	Node<T> * node = new Node<T>(item);
	Node<T> * current = head;
	if (head == NULL) { // sprawdzamy czy lista jest pusta
    	    head = node;
	    tail = node;
    	    node->next = NULL;
    	    node->prev = NULL;
	} else {
	    while (*node > *current) {
		current = current->next;
		if (!current) break;
	      }
    	    if ((*node) <= (*head)) { // jesli nowy element jest najmniejszy, glowa pokazuje na niego
		node -> next = head;
		head->prev = node;
		head = node;
		node->prev = NULL;
	    } else {
		if (current) { // jesli nie jestesmy na koncu wstawiamy node przed current
		    node->prev = current->prev;
		    node->next = current;
		    current->prev->next = node;
    		    current->prev = node;
		} else { //dodajemy na koncu
		    tail->next = node;
		    node->prev = tail;
		    node->next = NULL;
		    tail = node;
		}
	    }
	}
	c++;
	return *this;
    }
